#!/bin/bash
#
# rawtherapee-install:rawtherapee-install.sh
#
# Install last version of rawtherapee (using AppImage format) and start application
# (menu creation is handled by AppImage at first run)
#
set -u
set -o pipefail

declare -gr DEFAULT_APPIMAGES_INSTALL_COMMON_HOME='/usr/local/share/appimages'
declare -gr DEFAULT_APPIMAGES_INSTALL_ALL_USERS_RAWTHERAPEE_HOME="${DEFAULT_APPIMAGES_INSTALL_COMMON_HOME}/rawtherapee"

declare -gr DEFAULT_RAWTHERAPEE_DOWNLOAD_URL='https://rawtherapee.com/downloads/' # Web URL
declare -gr DEFAULT_RAWTHERAPEE_DOWNLOAD_BASE_URL='https://rawtherapee.com/shared/builds/linux/' # Since 2022-01

function usage {
  cat <<EOF
Install RawTherapee appimage

Usage:

  $( basename "$0" ) [--list]
  [sudo] $( basename "$0" ) [--force] [--home INSTALLATION_DIR] [--url URL]
  $( basename "$0" ) --install-user [--force] [--home INSTALLATION_DIR] [--url URL]
  $( basename "$0" ) --install-all-users [--force] [--home INSTALLATION_DIR] [--url URL]

Actions:

  -h,   --help      Display this message and exit.
  -l,   --list      List url of available versions.
  -i,   --install   (default action)
                    Install for current user only or for all users depend of if you use 'sudo' or not
  -U,   --install-user
                    Install for current user only
  -A,   --install-all-users
                    Install for all users (installer will use sudo if needed)

Options:

  -F,   --force     Force download even if file already present. Allow to reinstall existing
                    version of RawTherapee
        --home DIR  Set installation directory
                    (default value: '$( get_default_appimages_install_home )')
        --url URL   Set RawTherapee download page URL
                    (default value: '${DEFAULT_RAWTHERAPEE_DOWNLOAD_BASE_URL}')
        --debug
                    Add some traces

If you have any issues, check: ${DEFAULT_RAWTHERAPEE_DOWNLOAD_URL}

EOF
}

#
# P_param : get parameter value, if not exist call usage and return error.
#
# Usage:
#   param_xxx="$( P_param "$@" )" && shift || return $?
#
# Require:
#   - param variable with current parameter name
#   - usage function (use to display usage)
#
function P_param {
  if [ "$#" -gt 0 ] ; then
    echo "$1"
  else
    echofatal "Value required for '${param}'"
    usage >&2
    return 1
  fi
}

function echoinfo {
  echo >&2 '* [INFO]: ' "$@"
}

function echofatal {
  echo >&2 '*** [FATAL]: ' "$@"
}

function set_install_mode {
  local -r action="$1"

  case "${action}" in
    '--install-user'     ) MODE='user' ;;
    '--install-all-users') MODE='system' ;;

    '--install'|'')
      if [[ $(id -u) -ne 0 ]] ; then
        MODE='user'
      else
        MODE='system'
      fi
    ;;

    *)
      echofatal "Internal error: set_install_mode '${action}'"
      exit 100
    ;;
  esac

  if [ "${MODE}" = 'system' ] ; then
    SUDO_IF_NEEDED='sudo'
  else
    SUDO_IF_NEEDED=''
  fi
}

function fix_install_mode_if_needed {
  if [ "${MODE}" = 'user' ] && [ -d "${DEFAULT_APPIMAGES_INSTALL_ALL_USERS_RAWTHERAPEE_HOME}" ]; then
    while true; do
      # Getting LC_MESSAGES values into variables
      # shellcheck disable=SC2046 # Intended IFS splitting
      IFS='
' set -- $(locale LC_MESSAGES)

      yesexpr="$1"
      noexpr="$2"
      yesstr="$3"
      nostr="$4"
      # shellcheck disable=SC2034
      messages_codeset="$5" # unused here, but kept as documentation

      # Display Yes / No ? prompt into locale
      cat <<EOF >&2
* WARNING: Directory '${DEFAULT_APPIMAGES_INSTALL_ALL_USERS_RAWTHERAPEE_HOME}' found, but you are using 'single user' mode.
Do you wan't to continue and install RawTherapee for current user (${USER}) only ? (${yesstr} / ${nostr})
EOF

      # Read answer
      read -r yn

      # Test answer
      case "$yn" in
        # match only work with the character class from the expression
        "${yesexpr##^}"|"${yesstr}")
          echoinfo 'Use --install-user'
          break
          ;;

        "${noexpr##^}"|"${nostr}")
          MODE='system'
          echoinfo 'Use --install-all-users'
          break
          ;;

        *)
          echoinfo "Waiting for: ${yesexpr##^} / ${yesstr}  or ${noexpr##^} / ${nostr}"
          ;;
      esac
    done
  fi

  if [ "${MODE}" = 'system' ] ; then
    SUDO_IF_NEEDED='sudo'
  else
    SUDO_IF_NEEDED=''
  fi
}

function compute_DEFAULT_APPIMAGES_INSTALL_HOME {
  if [ "${MODE}" = 'system' ] ; then
    echo "${DEFAULT_APPIMAGES_INSTALL_ALL_USERS_RAWTHERAPEE_HOME}"
  else
    echo "${HOME}/Apps/rawtherapee"
  fi
}

# Verify magic number for Linux binaries
function check_binary {
  local -r file="$1"

  local filerealpath=
  filerealpath="$( realpath "${file}" )"

  local magic=
  magic="$( hexdump -n 4  -v -e '/1 "%02X "' "${file}" )"

  if [ "${magic}" = '7F 45 4C 46 ' ] ; then
    ${SUDO_IF_NEEDED} chmod -v +x "${file}" || exit 1
    echo "'${file}' is ready: (${filerealpath})"
  else
    echofatal "'${file}' look suspicious...  (MAGIC='${magic}'): (${filerealpath})"
    exit 1
  fi
}

function find_last_version_url {
  find_version_url | head -n 1
  # Expected result:
  #     https://rawtherapee.com/./releases_head/linux/RawTherapee-releases-5.6-20190420.AppImage

  return 0
}

function do_curl {
  local -r curl_url="$1"
  local -a curl_params=()

  if [ "${DEBUG:-}" != 'true' ] ; then
    curl_params+=(-s) # --silent
  fi
  curl_params+=(-S) # --show-error

  if [ "${DEBUG:-}" = 'true' ] ; then
    set -x
  fi
  curl "${curl_params[@]}" "${curl_url}" || return $?
}

function do_curl_and_filter {
  local -r curl_url="$1"

  local _page__content_=
  _page__content_="$( do_curl "${curl_url}" )" || return $?

  # Get link to AppImage files only
  _page__content_="$(
    grep ' href="RawTherapee' <<<"${_page__content_}" |
      sed -E 's,(.*) href="RawTherapee(.*).AppImage"(.*),RawTherapee\2.AppImage,' |
      grep -v ' href="RawTherapee'
   )"

   # Reorder according to version number
  sed -E 's,(.*)([0-9]+\.[0-9]+)(.*),\2#\1\2\3,' <<<"${_page__content_}" |
    sort --reverse |
    cut -d'#' -f2-
}

function find_version_url {
  local _files_=
  _files_="$( do_curl_and_filter "${download_url}" )" || return $?
  local -r _files_

  local _file_=
  if [ "${DEBUG:-}" = 'true' ] ; then
    while read -r _file_ ; do
      if [ -n "${_file_}" ] ; then
        echoinfo "FOUND: ${download_url}${_file_}"
      fi
    done <<<"${_files_}"
  fi

  while read -r _file_ ; do
    if [ -n "${_file_}" ] ; then
      echo "${download_url}${_file_}"
    fi
  done <<<"${_files_}"
}

function download_and_install {
  local -r bin_filename="$1"
  local -r bin_url="$2"

  cat <<EOF
  ${SUDO_IF_NEEDED} wget "${bin_url}" -O "${bin_filename}"
EOF
  ${SUDO_IF_NEEDED} wget "${bin_url}" -O "${bin_filename}" || exit 1
  check_binary "${bin_filename}"
}

function prerequisites {
  local -r install_home="$1"

  ${SUDO_IF_NEEDED} mkdir -vp "${install_home}" || exit 1
}

function download_and_install_if_needed {
  local last_url=
  local local_bin=

  local -r appimages_install_home="${param_appimages_install_home:-$( compute_DEFAULT_APPIMAGES_INSTALL_HOME )}"

  prerequisites "${appimages_install_home}"

  echo "Installation folder: '${appimages_install_home}'"

  last_url="$( find_last_version_url )"
  if [ -z "${last_url}" ] ; then
    cat <<EOF >&2

*** Cannot compute URL for last version of RawTherapee.

Try to use --list

*** Error while parsing: "${download_url}"
$( curl --head "${download_url}" )
EOF
    exit 1
  fi

  echo "* Selected URL: ${last_url}"

  local_bin="${appimages_install_home}/$( basename "${last_url}" )"
  if [ -f "${local_bin}" ] ; then
    if [ "${param_force_menu}" != 'true' ] && [ "${param_force_download}" != 'true' ] ; then
      cat <<EOF >&2
File already there '${local_bin}'

You already have last version.

Use --force to reinstall or --force-menu to update menu.
EOF
      exit 0
    fi
  fi

  if [ ! -f "${local_bin}" ] || [ "${param_force_download}" = 'true' ] ; then
    download_and_install "${local_bin}" "${last_url}"
  fi

  "${local_bin}" &
}

function main {
  local action='install'
  local param=
  local param_appimages_install_home=
  local param_download_url=
  local param_force_download='false'
  local param_force_menu='false'

  set_install_mode ''

  while [ "$#" -gt 0 ] ; do
    param="$1"
    shift

    case "${param}" in
      '-h'|'--help')
        usage
        return 0
        ;;

      '-l'|'--list'             ) action='--list' ;;
      '-i'|'--install'          ) action='--install' ;;
      '-U'|'--install-user'     ) action='--install-user' ;;
      '-A'|'--install-all-users') action='--install-all-users' ;;

      '--home') param_appimages_install_home="$( P_param "$@" )" && shift || return $? ;;
      '--url' ) param_download_url="$( P_param "$@" )" && shift || return $? ;;

      '-F'|'--force'    ) param_force_menu='true' ; param_force_download='true' ;;
      '--force-download') param_force_download='true' ;;
      '--force-menu'    ) param_force_menu='true' ;;
      '--debug'         ) DEBUG='true' ;;

      *)
        echofatal "Unknown option: '${param}'"
        usage >&2
        return 1
        ;;
    esac
  done

  local -r download_url="${param_download_url:-${DEFAULT_RAWTHERAPEE_DOWNLOAD_BASE_URL}}"

  case "${action}" in
    '--list')
      find_version_url || return $?
      ;;

    '--install-user'|'--install-all-users')
      set_install_mode "${action}" || return $?
      fix_install_mode_if_needed || return $?
      download_and_install_if_needed || return $?
      ;;

    '--install'|*)
      set_install_mode '' || return $?
      fix_install_mode_if_needed || return $?
      download_and_install_if_needed || return $?
      ;;
  esac
}

main "$@" || exit $?
